@@ -93,7 +93,7 @@ usage (int status)
 Usage: %s [OPTION] [FILE]...\n\
 "),
 	      program_name);
-      printf (_("\
+      fputs (_("\
 Concatenate FILE(s), or standard input, to standard output.\n\
 \n\
   -A, --show-all           equivalent to -vET\n\
@@ -102,20 +102,24 @@ Concatenate FILE(s), or standard input, to standard output.\n\
   -E, --show-ends          display $ at end of each line\n\
   -n, --number             number all output lines\n\
   -s, --squeeze-blank      never more than one single blank line\n\
+"), stdout);
+      fputs (_("\
   -t                       equivalent to -vT\n\
   -T, --show-tabs          display TAB characters as ^I\n\
   -u                       (ignored)\n\
   -v, --show-nonprinting   use ^ and M- notation, except for LFD and TAB\n\
-      --help               display this help and exit\n\
-      --version            output version information and exit\n\
+"), stdout);
+      fputs (HELP_OPTION_DESCRIPTION, stdout);
+      fputs (VERSION_OPTION_DESCRIPTION, stdout);
+      fputs (_("\
 \n\
 With no FILE, or when FILE is -, read standard input.\n\
-"));
+"), stdout);
 #if O_BINARY
-      printf (_("\
+      fputs (_("\
 \n\
   -B, --binary             use binary writes to the console device.\n\n\
-"));
+"), stdout);
 #endif
       puts (_("\nReport bugs to <bug-textutils@gnu.org>."));
     }
@@ -259,7 +263,7 @@ cat (
 	      unsigned char *wp = outbuf;
 	      do
 		{
-		  if (full_write (STDOUT_FILENO, wp, outsize) < 0)
+		  if (full_write (STDOUT_FILENO, wp, outsize) != outsize)
 		    error (EXIT_FAILURE, errno, _("write error"));
 		  wp += outsize;
 		}
@@ -312,7 +316,7 @@ cat (
 		{
 		  int n_write = bpout - outbuf;
 
-		  if (full_write (STDOUT_FILENO, outbuf, n_write) < 0)
+		  if (full_write (STDOUT_FILENO, outbuf, n_write) != n_write)
 		    error (EXIT_FAILURE, errno, _("write error"));
 		  bpout = outbuf;
 		}
