@@ -1,5 +1,5 @@
 /* tail -- output the last part of file(s)
-   Copyright (C) 1989, 90, 91, 1995-2001 Free Software Foundation, Inc.
+   Copyright (C) 1989, 90, 91, 1995-2002 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -35,6 +35,8 @@
 #include "closeout.h"
 #include "argmatch.h"
 #include "error.h"
+#include "human.h"
+#include "posixver.h"
 #include "safe-read.h"
 #include "xstrtol.h"
 
@@ -242,52 +244,63 @@ Print the last %d lines of each FILE to standard output.\n\
 With more than one FILE, precede each with a header giving the file name.\n\
 With no FILE, or when FILE is -, read standard input.\n\
 \n\
+"), DEFAULT_N_LINES);
+     fputs (_("\
+Mandatory arguments to long options are mandatory for short options too.\n\
+"), stdout);
+     fputs (_("\
       --retry              keep trying to open a file even if it is\n\
                            inaccessible when tail starts or if it becomes\n\
                            inaccessible later -- useful only with -f\n\
   -c, --bytes=N            output the last N bytes\n\
+"), stdout);
+     fputs (_("\
   -f, --follow[={name|descriptor}]\n\
                            output appended data as the file grows;\n\
                            -f, --follow, and --follow=descriptor are\n\
                            equivalent\n\
   -F                       same as --follow=name --retry\n\
+"), stdout);
+     printf (_("\
   -n, --lines=N            output the last N lines, instead of the last %d\n\
       --max-unchanged-stats=N\n\
                            with --follow=name, reopen a FILE which has not\n\
                            changed size after N (default %d) iterations\n\
                            to see if it has been unlinked or renamed\n\
                            (this is the usual case of rotated log files)\n\
+"),
+	     DEFAULT_N_LINES,
+	     DEFAULT_MAX_N_UNCHANGED_STATS_BETWEEN_OPENS
+	     );
+     fputs (_("\
       --pid=PID            with -f, terminate after process ID, PID dies\n\
   -q, --quiet, --silent    never output headers giving file names\n\
   -s, --sleep-interval=S   with -f, each iteration lasts approximately S\n\
                            (default 1) seconds\n\
   -v, --verbose            always output headers giving file names\n\
-      --help               display this help and exit\n\
-      --version            output version information and exit\n\
+"), stdout);
+     fputs (HELP_OPTION_DESCRIPTION, stdout);
+     fputs (VERSION_OPTION_DESCRIPTION, stdout);
+     fputs (_("\
 \n\
-"),
-	      DEFAULT_N_LINES, DEFAULT_N_LINES,
-	      DEFAULT_MAX_N_UNCHANGED_STATS_BETWEEN_OPENS
-	      );
-      printf (_("\
 If the first character of N (the number of bytes or lines) is a `+',\n\
 print beginning with the Nth item from the start of each file, otherwise,\n\
 print the last N items in the file.  N may have a multiplier suffix:\n\
-b for 512, k for 1024, m for 1048576 (1 Meg).  A first OPTION of -VALUE\n\
-or +VALUE is treated like -n VALUE or -n +VALUE unless VALUE has one of\n\
-the [bkm] suffix multipliers, in which case it is treated like -c VALUE\n\
-or -c +VALUE.  Warning: a first option of +VALUE is obsolescent, and support\n\
-for it will be withdrawn.\n\
+b for 512, k for 1024, m for 1048576 (1 Meg).\n\
 \n\
+"), stdout);
+     fputs (_("\
 With --follow (-f), tail defaults to following the file descriptor, which\n\
 means that even if a tail'ed file is renamed, tail will continue to track\n\
-its end.  This default behavior is not desirable when you really want to\n\
+its end.  \
+"), stdout);
+     fputs (_("\
+This default behavior is not desirable when you really want to\n\
 track the actual name of the file, not the file descriptor (e.g., log\n\
 rotation).  Use --follow=name in that case.  That causes tail to track the\n\
 named file by reopening it periodically to see if it has been removed and\n\
 recreated by some other program.\n\
-\n\
-"));
+"), stdout);
       puts (_("\nReport bugs to <bug-textutils@gnu.org>."));
     }
   exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
@@ -310,7 +323,6 @@ static void
 xwrite (int fd, char *const buffer, size_t n_bytes)
 {
   assert (fd == STDOUT_FILENO);
-  assert (n_bytes >= 0);
   if (n_bytes > 0 && fwrite (buffer, 1, n_bytes, stdout) == 0)
     error (EXIT_FAILURE, errno, _("write error"));
 }
@@ -362,21 +374,61 @@ dump_remainder (const char *pretty_filename, int fd, off_t n_bytes)
   return n_written;
 }
 
+/* Call lseek with the specified arguments, where file descriptor FD
+   corresponds to the file, FILENAME.
+   Give a diagnostic and exit nonzero if lseek fails.  */
+
+static void
+xlseek (int fd, off_t offset, int whence, char const *filename)
+{
+  off_t new_offset = lseek (fd, offset, whence);
+  char buf[LONGEST_HUMAN_READABLE + 1];
+  char *s;
+  char const *sign;
+
+  if (0 <= new_offset)
+    return;
+
+  sign = offset < 0 ? "-" : "";
+  if (offset < 0)
+    offset = -offset;
+
+  s = human_readable ((uintmax_t) offset, buf, 1, 1);
+  switch (whence)
+    {
+    case SEEK_SET:
+      error (1, errno, _("%s: cannot seek to offset %s%s"),
+	     filename, sign, s);
+      break;
+    case SEEK_CUR:
+      error (1, errno, _("%s: cannot seek to relative offset %s%s"),
+	     filename, sign, s);
+      break;
+    case SEEK_END:
+      error (1, errno, _("%s: cannot seek to end-relative offset %s%s"),
+	     filename, sign, s);
+      break;
+    default:
+      abort ();
+    }
+}
+
 /* Print the last N_LINES lines from the end of file FD.
    Go backward through the file, reading `BUFSIZ' bytes at a time (except
    probably the first), until we hit the start of the file or have
    read NUMBER newlines.
+   START_POS is the starting position of the read pointer for the file
+   associated with FD (may be nonzero).
    FILE_LENGTH is the length of the file (one more than the offset of the
    last byte of the file).
    Return 0 if successful, 1 if an error occurred.  */
 
 static int
 file_lines (const char *pretty_filename, int fd, long int n_lines,
-	    off_t file_length)
+	    off_t start_pos, off_t file_length)
 {
   char buffer[BUFSIZ];
   int bytes_read;
-  int i;			/* Index into `buffer' for scanning.  */
   off_t pos = file_length;
 
   if (n_lines == 0)
@@ -384,14 +436,13 @@ file_lines (const char *pretty_filename, int fd, long int n_lines,
 
   /* Set `bytes_read' to the size of the last, probably partial, buffer;
      0 < `bytes_read' <= `BUFSIZ'.  */
-  bytes_read = pos % BUFSIZ;
+  bytes_read = (pos - start_pos) % BUFSIZ;
   if (bytes_read == 0)
     bytes_read = BUFSIZ;
   /* Make `pos' a multiple of `BUFSIZ' (0 if the file is short), so that all
      reads will be on block boundaries, which might increase efficiency.  */
   pos -= bytes_read;
-  /* FIXME: check lseek return value  */
-  lseek (fd, pos, SEEK_SET);
+  xlseek (fd, pos, SEEK_SET, pretty_filename);
   bytes_read = safe_read (fd, buffer, bytes_read);
   if (bytes_read == -1)
     {
@@ -405,6 +456,7 @@ file_lines (const char *pretty_filename, int fd, long int n_lines,
 
   do
     {
+      int i;			/* Index into `buffer' for scanning.  */
       /* Scan backward, counting the newlines in this bufferfull.  */
       for (i = bytes_read - 1; i >= 0; i--)
 	{
@@ -421,17 +473,15 @@ file_lines (const char *pretty_filename, int fd, long int n_lines,
 	    }
 	}
       /* Not enough newlines in that bufferfull.  */
-      if (pos == 0)
+      if (pos == start_pos)
 	{
 	  /* Not enough lines in the file; print the entire file.  */
-	  /* FIXME: check lseek return value  */
-	  lseek (fd, (off_t) 0, SEEK_SET);
+	  xlseek (fd, start_pos, SEEK_SET, pretty_filename);
 	  dump_remainder (pretty_filename, fd, file_length);
 	  return 0;
 	}
       pos -= BUFSIZ;
-      /* FIXME: check lseek return value  */
-      lseek (fd, pos, SEEK_SET);
+      xlseek (fd, pos, SEEK_SET, pretty_filename);
     }
   while ((bytes_read = safe_read (fd, buffer, BUFSIZ)) > 0);
 
@@ -687,7 +737,7 @@ start_bytes (const char *pretty_filename, int fd, off_t n_bytes)
 
 /* Skip N_LINES lines at the start of file or pipe FD, and print
    any extra characters that were read beyond that.
-   Return 1 on error, 0 if ok.  */
+   Return 1 on error, 0 if ok, -1 if EOF.  */
 
 static int
 start_lines (const char *pretty_filename, int fd, long int n_lines)
@@ -696,6 +746,9 @@ start_lines (const char *pretty_filename, int fd, long int n_lines)
   int bytes_read = 0;
   int bytes_to_skip = 0;
 
+  if (n_lines == 0)
+    return 0;
+
   while (n_lines && (bytes_read = safe_read (fd, buffer, BUFSIZ)) > 0)
     {
       bytes_to_skip = 0;
@@ -703,16 +756,24 @@ start_lines (const char *pretty_filename, int fd, long int n_lines)
 	if (buffer[bytes_to_skip++] == '\n' && --n_lines == 0)
 	  break;
     }
+
+  /* EOF */
+  if (bytes_read == 0)
+    return -1;
+
+  /* error */
   if (bytes_read == -1)
     {
       error (0, errno, "%s", pretty_filename);
       return 1;
     }
-  else if (bytes_to_skip < bytes_read)
+
+  if (bytes_to_skip < bytes_read)
     {
       xwrite (STDOUT_FILENO, &buffer[bytes_to_skip],
 	      bytes_read - bytes_to_skip);
     }
+
   return 0;
 }
 
@@ -836,8 +897,7 @@ recheck (struct File_spec *f)
       f->n_unchanged_stats = 0;
       f->n_consecutive_size_changes = 0;
       f->ignore = 0;
-      /* FIXME: check lseek return value  */
-      lseek (f->fd, f->size, SEEK_SET);
+      xlseek (f->fd, f->size, SEEK_SET, pretty_name (f));
     }
 }
 
@@ -937,8 +997,8 @@ tail_forever (struct File_spec *f, int nfiles)
 	    {
 	      error (0, 0, _("%s: file truncated"), pretty_name (&f[i]));
 	      last = i;
-	      /* FIXME: check lseek return value  */
-	      lseek (f[i].fd, stats.st_size, SEEK_SET);
+	      xlseek (f[i].fd, (off_t) stats.st_size, SEEK_SET,
+		      pretty_name (&f[i]));
 	      f[i].size = stats.st_size;
 	      continue;
 	    }
@@ -1000,8 +1060,7 @@ tail_bytes (const char *pretty_filename, int fd, off_t n_bytes)
     {
       if (S_ISREG (stats.st_mode))
 	{
-	  /* FIXME: check lseek return value  */
-	  lseek (fd, n_bytes, SEEK_CUR);
+	  xlseek (fd, n_bytes, SEEK_CUR, pretty_filename);
 	}
       else if (start_bytes (pretty_filename, fd, n_bytes))
 	{
@@ -1036,15 +1095,13 @@ tail_bytes (const char *pretty_filename, int fd, off_t n_bytes)
 		 more bytes than have been requested.  So reposition the
 		 file pointer to the incoming current position and print
 		 everything after that.  */
-	      /* FIXME: check lseek return value  */
-	      lseek (fd, current_pos, SEEK_SET);
+	      xlseek (fd, current_pos, SEEK_SET, pretty_filename);
 	    }
 	  else
 	    {
 	      /* There are more bytes remaining than were requested.
 		 Back up.  */
-	      /* FIXME: check lseek return value  */
-	      lseek (fd, -n_bytes, SEEK_END);
+	      xlseek (fd, -n_bytes, SEEK_END, pretty_filename);
 	    }
 	  dump_remainder (pretty_filename, fd, n_bytes);
 	}
@@ -1061,7 +1118,6 @@ static int
 tail_lines (const char *pretty_filename, int fd, long int n_lines)
 {
   struct stat stats;
-  off_t length;
 
   /* We need binary input, since `tail' relies on `lseek' and byte counts,
      while binary output will preserve the style (Unix/DOS) of text file.  */
@@ -1075,23 +1131,26 @@ tail_lines (const char *pretty_filename, int fd, long int n_lines)
 
   if (from_start)
     {
-      if (start_lines (pretty_filename, fd, n_lines))
+      int t;
+      if ((t = start_lines (pretty_filename, fd, n_lines)) < 0)
+	return 0;
+      if (t)
 	return 1;
       dump_remainder (pretty_filename, fd, COPY_TO_EOF);
     }
   else
     {
-      /* Use file_lines only if FD refers to a regular file with
-         its file pointer positioned at beginning of file.  */
-      /* FIXME: adding the lseek conjunct is a kludge.
-	 Once there's a reasonable test suite, fix the true culprit:
-	 file_lines.  file_lines shouldn't presume that the input
-	 file pointer is initially positioned to beginning of file.  */
+      off_t length;
+      off_t start_pos;
+
+      /* Use file_lines only if FD refers to a regular file for
+	 which lseek (... SEEK_END) works.  */
       if (S_ISREG (stats.st_mode)
-	  && lseek (fd, (off_t) 0, SEEK_CUR) == (off_t) 0)
+	  && (start_pos = lseek (fd, (off_t) 0, SEEK_CUR)) != -1
+	  && start_pos < (length = lseek (fd, (off_t) 0, SEEK_END)))
 	{
-	  length = lseek (fd, (off_t) 0, SEEK_END);
-	  if (length != 0 && file_lines (pretty_filename, fd, n_lines, length))
+	  if (length != 0 && file_lines (pretty_filename, fd, n_lines,
+					 start_pos, length))
 	    return 1;
 	}
       else
@@ -1120,7 +1179,6 @@ static int
 tail_file (struct File_spec *f, off_t n_units)
 {
   int fd, errors;
-  struct stat stats;
 
   int is_stdin = (STREQ (f->name, "-"));
 
@@ -1156,8 +1214,9 @@ tail_file (struct File_spec *f, off_t n_units)
       errors = tail (pretty_name (f), fd, n_units);
       if (forever)
 	{
+	  struct stat stats;
+
 	  f->errnum = 0;
-	  /* FIXME: duplicate code */
 	  if (fstat (fd, &stats) < 0)
 	    {
 	      errors = 1;
@@ -1219,6 +1278,7 @@ parse_obsolescent_option (int argc, const char *const *argv,
   const char *p = argv[1];
   const char *n_string = NULL;
   const char *n_string_end;
+  bool obsolete_usage;
 
   int t_from_start;
   int t_count_lines;
@@ -1230,9 +1290,13 @@ parse_obsolescent_option (int argc, const char *const *argv,
   if (argc < 2)
     return 0;
 
-  /* If P starts with `+', `-N' (where N is a digit), or `-l',
-     then it is obsolescent.  Return zero otherwise.  */
-  if ( ! (p[0] == '+' || (p[0] == '-' && (p[1] == 'l' || ISDIGIT (p[1])))) )
+  obsolete_usage = (posix2_version () < 200112);
+
+  /* If P starts with `+' and the POSIX version predates 1003.1-2001,
+     or if P starts with `-N' (where N is a digit), or `-l', then it
+     is obsolescent.  Return zero otherwise.  */
+  if (! ((p[0] == '+' && obsolete_usage)
+	 || (p[0] == '-' && (p[1] == 'l' || ISDIGIT (p[1])))))
     return 0;
 
   if (*p == '+')
@@ -1326,11 +1390,9 @@ parse_obsolescent_option (int argc, const char *const *argv,
     {
       if (argc > 3)
 	{
-	  int posix_pedantic = (getenv ("POSIXLY_CORRECT") != NULL);
-
 	  /* When POSIXLY_CORRECT is set, enforce the `at most one
 	     file argument' requirement.  */
-	  if (posix_pedantic)
+	  if (getenv ("POSIXLY_CORRECT"))
 	    {
 	      error (0, 0, _("\
 too many arguments;  When using tail's obsolescent option syntax (%s)\n\
@@ -1348,6 +1410,14 @@ option instead."), argv[1]);
 #endif
 	}
 
+      if (! obsolete_usage)
+	{
+	  error (0, 0, _("`%s' option is obsolete; use `%s-%c %.*s'"),
+		 argv[1], t_forever ? " -f" : "", t_count_lines ? 'n' : 'c',
+		 (int) (n_string_end - n_string), n_string);
+	  usage (EXIT_FAILURE);
+	}
+
       /* Set globals.  */
       from_start = t_from_start;
       count_lines = t_count_lines;
@@ -1378,6 +1448,7 @@ parse_options (int argc, char **argv,
 	  forever = 1;
 	  follow_mode = Follow_name;
 	  reopen_inaccessible_files = 1;
+	  break;
 
 	case 'c':
 	case 'n':
@@ -1529,12 +1600,11 @@ main (int argc, char **argv)
   have_read_stdin = 0;
 
   {
-    int found_obsolescent;
     int fail;
-    found_obsolescent = parse_obsolescent_option (argc,
-						  (const char *const *) argv,
-						  &n_units, &fail);
-    if (found_obsolescent)
+
+    if (parse_obsolescent_option (argc,
+				  (const char *const *) argv,
+				  &n_units, &fail))
       {
 	if (fail)
 	  exit (EXIT_FAILURE);
